/*
 * Copyright (C) 2006 The Android Open Source Project
 * Copyright (c) 2011, Code Aurora Forum. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package android.webkit;

import android.content.Context;
import android.net.http.AndroidHttpClient;
import android.net.http.Headers;
import android.os.FileUtils;
import android.os.SystemProperties;
import android.util.Log;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.List;
import java.util.Map;

import org.bouncycastle.crypto.Digest;
import org.bouncycastle.crypto.digests.SHA1Digest;

import java.util.ArrayList;
import java.util.Collections;
import java.io.Serializable;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.lang.Integer;
import java.lang.Long;
import java.lang.Thread;
import java.lang.Runnable;

/**
 * The class CacheManager provides the persistent cache of content that is
 * received over the network. The component handles parsing of HTTP headers and
 * utilizes the relevant cache headers to determine if the content should be
 * stored and if so, how long it is valid for. Network requests are provided to
 * this component and if they can not be resolved by the cache, the HTTP headers
 * are attached, as appropriate, to the request for revalidation of content. The
 * class also manages the cache size.
 */
public final class CacheManager {

    private static final String LOGTAG = "cache";

    static final String HEADER_KEY_IFMODIFIEDSINCE = "if-modified-since";
    static final String HEADER_KEY_IFNONEMATCH = "if-none-match";

    private static final String NO_STORE = "no-store";
    private static final String NO_CACHE = "no-cache";
    private static final String MAX_AGE = "max-age";
    private static final String MANIFEST_MIME = "text/cache-manifest";

    private static final long CACHE_THRESHOLD_DEF = 6 * 1024 * 1024;
    private static final long CACHE_TRIM_AMOUNT_DEF = 2 * 1024 * 1024;

    private static long CACHE_THRESHOLD = CACHE_THRESHOLD_DEF;
    private static long CACHE_TRIM_AMOUNT = CACHE_TRIM_AMOUNT_DEF;

    // Limit the maximum cache file size to half of the normal capacity
    static long CACHE_MAX_SIZE = (CACHE_THRESHOLD_DEF - CACHE_TRIM_AMOUNT_DEF) / 2;

    private static boolean mDisabled;

    // Reference count the enable/disable transaction
    private static int mRefCount;

    // trimCacheIfNeeded() is called when a page is fully loaded. But JavaScript
    // can load the content, e.g. in a slideshow, continuously, so we need to
    // trim the cache on a timer base too. endCacheTransaction() is called on a 
    // timer base. We share the same timer with less frequent update.
    private static int mTrimCacheCount = 0;
    private static final int TRIM_CACHE_INTERVAL = 5;

    private static WebViewDatabase mDataBase;
    private static File mBaseDir;
    
    // Flag to clear the cache when the CacheManager is initialized
    private static boolean mClearCacheOnInit = false;

    /**
     * This class represents a resource retrieved from the HTTP cache.
     * Instances of this class can be obtained by invoking the
     * CacheManager.getCacheFile() method.
     */
    public static class CacheResult {
        // these fields are saved to the database
        int httpStatusCode;
        long contentLength;
        long expires;
        String expiresString;
        String localPath;
        String lastModified;
        String etag;
        String mimeType;
        String location;
        String encoding;
        String contentdisposition;
        String crossDomain;
        int accessCounter;

        // these fields are NOT saved to the database
        InputStream inStream;
        OutputStream outStream;
        File outFile;

        public int getHttpStatusCode() {
            return httpStatusCode;
        }

        public long getContentLength() {
            return contentLength;
        }

        public String getLocalPath() {
            return localPath;
        }

        public long getExpires() {
            return expires;
        }

        public String getExpiresString() {
            return expiresString;
        }

        public String getLastModified() {
            return lastModified;
        }

        public String getETag() {
            return etag;
        }

        public String getMimeType() {
            return mimeType;
        }

        public String getLocation() {
            return location;
        }

        public String getEncoding() {
            return encoding;
        }

        public String getContentDisposition() {
            return contentdisposition;
        }

        // For out-of-package access to the underlying streams.
        public InputStream getInputStream() {
            return inStream;
        }

        public OutputStream getOutputStream() {
            return outStream;
        }

        // These fields can be set manually.
        public void setInputStream(InputStream stream) {
            this.inStream = stream;
        }

        public void setEncoding(String encoding) {
            this.encoding = encoding;
        }
    }

    /**
     * initialize the CacheManager. WebView should handle this for each process.
     * 
     * @param context The application context.
     */
    static void init(Context context) {
    	speculativeInit();
    	
        mDataBase = WebViewDatabase.getInstance(context.getApplicationContext());
        mBaseDir = new File(context.getCacheDir(), "webviewCache");
        if (createCacheDirectory() && mClearCacheOnInit) {
            removeAllCacheFiles();
            mClearCacheOnInit = false;
        }

        CACHE_THRESHOLD = SystemProperties.getLong("nw.cache.threshold", CACHE_THRESHOLD_DEF);
        CACHE_TRIM_AMOUNT = SystemProperties.getLong("nw.cache.trimamount", CACHE_TRIM_AMOUNT_DEF);
        CACHE_MAX_SIZE = (CACHE_THRESHOLD - CACHE_TRIM_AMOUNT) / 2;
    }
    
    /**
     * Create the cache directory if it does not already exist.
     * 
     * @return true if the cache directory didn't exist and was created.
     */
    static private boolean createCacheDirectory() {
        if (!mBaseDir.exists()) {
            if(!mBaseDir.mkdirs()) {
                Log.w(LOGTAG, "Unable to create webviewCache directory");
                return false;
            }
            FileUtils.setPermissions(
                    mBaseDir.toString(),
                    FileUtils.S_IRWXU|FileUtils.S_IRWXG|FileUtils.S_IXOTH,
                    -1, -1);
            // If we did create the directory, we need to flush 
            // the cache database. The directory could be recreated
            // because the system flushed all the data/cache directories
            // to free up disk space.
            // delete rows in the cache database
            WebViewWorker.getHandler().sendEmptyMessage(
                    WebViewWorker.MSG_CLEAR_CACHE);
            return true;
        }
        return false;
    }

    /**
     * get the base directory of the cache. With localPath of the CacheResult,
     * it identifies the cache file.
     * 
     * @return File The base directory of the cache.
     */
    public static File getCacheFileBaseDir() {
        return mBaseDir;
    }

    /**
     * set the flag to control whether cache is enabled or disabled
     * 
     * @param disabled true to disable the cache
     */
    static void setCacheDisabled(boolean disabled) {
        if (disabled == mDisabled) {
            return;
        }
        mDisabled = disabled;
        if (mDisabled) {
            removeAllCacheFiles();
        }
    }

    /**
     * get the state of the current cache, enabled or disabled
     * 
     * @return return if it is disabled
     */
    public static boolean cacheDisabled() {
        return mDisabled;
    }

    // only called from WebViewWorkerThread
    // make sure to call enableTransaction/disableTransaction in pair
    static boolean enableTransaction() {
        if (++mRefCount == 1) {
            mDataBase.startCacheTransaction();
            return true;
        }
        return false;
    }

    // only called from WebViewWorkerThread
    // make sure to call enableTransaction/disableTransaction in pair
    static boolean disableTransaction() {
        if (--mRefCount == 0) {
            mDataBase.endCacheTransaction();
            return true;
        }
        return false;
    }

    // only called from WebViewWorkerThread
    // make sure to call startTransaction/endTransaction in pair
    static boolean startTransaction() {
        return mDataBase.startCacheTransaction();
    }

    // only called from WebViewWorkerThread
    // make sure to call startTransaction/endTransaction in pair
    static boolean endTransaction() {
        boolean ret = mDataBase.endCacheTransaction();
        if (++mTrimCacheCount >= TRIM_CACHE_INTERVAL) {
            mTrimCacheCount = 0;
            trimCacheIfNeeded();
        }
        return ret;
    }

    // only called from WebCore Thread
    // make sure to call startCacheTransaction/endCacheTransaction in pair
    /**
     * @deprecated Always returns false.
     */
    @Deprecated
    public static boolean startCacheTransaction() {
        return false;
    }

    // only called from WebCore Thread
    // make sure to call startCacheTransaction/endCacheTransaction in pair
    /**
     * @deprecated Always returns false.
     */
    @Deprecated
    public static boolean endCacheTransaction() {
        return false;
    }

    /**
     * Given a url, returns the CacheResult if exists. Otherwise returns null.
     * If headers are provided and a cache needs validation,
     * HEADER_KEY_IFNONEMATCH or HEADER_KEY_IFMODIFIEDSINCE will be set in the
     * cached headers.
     * 
     * @return the CacheResult for a given url
     */
    public static CacheResult getCacheFile(String url,
            Map<String, String> headers) {
        return getCacheFile(url, 0, headers);
    }

    static CacheResult getCacheFile(String url, long postIdentifier,
            Map<String, String> headers) {
        if (mDisabled) {
            return null;
        }

        String databaseKey = getDatabaseKey(url, postIdentifier);

        CacheResult result = mDataBase.getCache(databaseKey);
        if (result != null) {
            if (result.contentLength == 0) {
                if (!checkCacheRedirect(result.httpStatusCode)) {
                    // this should not happen. If it does, remove it.
                    mDataBase.removeCache(databaseKey);
                    return null;
                }
            } else {
                File src = new File(mBaseDir, result.localPath);
                try {
                    // open here so that even the file is deleted, the content
                    // is still readable by the caller until close() is called
                    result.inStream = new FileInputStream(src);
                } catch (FileNotFoundException e) {
                    // the files in the cache directory can be removed by the
                    // system. If it is gone, clean up the database
                    mDataBase.removeCache(databaseKey);
                    return null;
                }
            }
        } else {
            return null;
        }

        if (speculativeEnabled()) {
            Log.d("speculative", "CacheManager: getCacheFile (in cache) | url="+url);
            // if url is speculatively loaded, return cache anyway
            if (inSpeculativeFinishedPool(url)) {
                return result;
            }
        }
        
        // null headers request coming from CACHE_MODE_CACHE_ONLY
        // which implies that it needs cache even it is expired.
        // negative expires means time in the far future.
        if (headers != null && result.expires >= 0
                && result.expires <= System.currentTimeMillis()) {
        	if (speculativeEnabled()) {
        		Log.d("speculative", "CacheManager: getCacheFile expire!!!! | url="+url);
        	}
            if (result.lastModified == null && result.etag == null) {
                return null;
            }
            // return HEADER_KEY_IFNONEMATCH or HEADER_KEY_IFMODIFIEDSINCE
            // for requesting validation
            if (result.etag != null) {
                headers.put(HEADER_KEY_IFNONEMATCH, result.etag);
            }
            if (result.lastModified != null) {
                headers.put(HEADER_KEY_IFMODIFIEDSINCE, result.lastModified);
            }
        }
        else if (headers != null && cacheExpireEnabled()) {
        	if (speculativeEnabled()) {
        		Log.d("speculative", "CacheManager: getCacheFile fresh, set to expire!!!! | url="+url);
        	}
            if (result.lastModified == null && result.etag == null) {
                return null;
            }
            if (result.etag != null) {
                headers.put(HEADER_KEY_IFNONEMATCH, result.etag);
            }
            if (result.lastModified != null) {
                headers.put(HEADER_KEY_IFMODIFIEDSINCE, result.lastModified);
            }
        }
        
        if (DebugFlags.CACHE_MANAGER) {
            Log.v(LOGTAG, "getCacheFile for url " + url);
        }

        return result;
    }

    /**
     * Given a url and its full headers, returns CacheResult if a local cache
     * can be stored. Otherwise returns null. The mimetype is passed in so that
     * the function can use the mimetype that will be passed to WebCore which
     * could be different from the mimetype defined in the headers.
     * forceCache is for out-of-package callers to force creation of a
     * CacheResult, and is used to supply surrogate responses for URL
     * interception.
     * @return CacheResult for a given url
     * @hide - hide createCacheFile since it has a parameter of type headers, which is
     * in a hidden package.
     */
    public static CacheResult createCacheFile(String url, int statusCode,
            Headers headers, String mimeType, boolean forceCache) {
        return createCacheFile(url, statusCode, headers, mimeType, 0,
                forceCache);
    }

    static CacheResult createCacheFile(String url, int statusCode,
            Headers headers, String mimeType, long postIdentifier,
            boolean forceCache) {
        if (!forceCache && mDisabled) {
            return null;
        }

        String databaseKey = getDatabaseKey(url, postIdentifier);

        // according to the rfc 2616, the 303 response MUST NOT be cached.
        if (statusCode == 303) {
            // remove the saved cache if there is any
            mDataBase.removeCache(databaseKey);
            return null;
        }

        // like the other browsers, do not cache redirects containing a cookie
        // header.
        if (checkCacheRedirect(statusCode) && !headers.getSetCookie().isEmpty()) {
            // remove the saved cache if there is any
            mDataBase.removeCache(databaseKey);
            return null;
        }

        CacheResult ret = parseHeaders(statusCode, headers, mimeType);
        if (ret == null) {
            // this should only happen if the headers has "no-store" in the
            // cache-control. remove the saved cache if there is any
            mDataBase.removeCache(databaseKey);
        } else {
        	if (speculativeEnabled()) {
        		Log.d("speculative", "CacheManager: createCacheFile url="+url);
        	}
            setupFiles(databaseKey, ret);
            try {
                ret.outStream = new FileOutputStream(ret.outFile);
            } catch (FileNotFoundException e) {
                // This can happen with the system did a purge and our
                // subdirectory has gone, so lets try to create it again
                if (createCacheDirectory()) {
                    try {
                        ret.outStream = new FileOutputStream(ret.outFile);
                    } catch  (FileNotFoundException e2) {
                        // We failed to create the file again, so there
                        // is something else wrong. Return null.
                        return null;
                    }
                } else {
                    // Failed to create cache directory
                    return null;
                }
            }
            ret.mimeType = mimeType;
        }

        return ret;
    }

    /**
     * Save the info of a cache file for a given url to the CacheMap so that it
     * can be reused later
     */
    public static void saveCacheFile(String url, CacheResult cacheRet) {
        saveCacheFile(url, 0, cacheRet);
    }

    static void saveCacheFile(String url, long postIdentifier,
            CacheResult cacheRet) {
        try {
            cacheRet.outStream.close();
        } catch (IOException e) {
            return;
        }

        if (!cacheRet.outFile.exists()) {
            // the file in the cache directory can be removed by the system
            return;
        }

        boolean redirect = checkCacheRedirect(cacheRet.httpStatusCode);
        if (redirect) {
            // location is in database, no need to keep the file
            cacheRet.contentLength = 0;
            cacheRet.localPath = "";
        }
        if ((redirect || cacheRet.contentLength == 0)
                && !cacheRet.outFile.delete()) {
            Log.e(LOGTAG, cacheRet.outFile.getPath() + " delete failed.");
        }
        if (cacheRet.contentLength == 0) {
            return;
        }

        mDataBase.addCache(getDatabaseKey(url, postIdentifier), cacheRet);

        if (DebugFlags.CACHE_MANAGER) {
            Log.v(LOGTAG, "saveCacheFile for url " + url);
        }
        
        if (speculativeEnabled()) {
        	Log.d("speculative", "CacheManager: ---- saveCacheFile ---- | url="+url);
        }
    }

    static boolean cleanupCacheFile(CacheResult cacheRet) {
        try {
            cacheRet.outStream.close();
        } catch (IOException e) {
            return false;
        }
        return cacheRet.outFile.delete();
    }

    /**
     * remove all cache files
     * 
     * @return true if it succeeds
     */
    static boolean removeAllCacheFiles() {
        // Note, this is called before init() when the database is
        // created or upgraded.
        if (mBaseDir == null) {
            // Init() has not been called yet, so just flag that
            // we need to clear the cache when init() is called.
            mClearCacheOnInit = true;
            return true;
        }
        // delete rows in the cache database
        WebViewWorker.getHandler().sendEmptyMessage(
                WebViewWorker.MSG_CLEAR_CACHE);
        // delete cache files in a separate thread to not block UI.
        final Runnable clearCache = new Runnable() {
            public void run() {
                // delete all cache files
                try {
                    String[] files = mBaseDir.list();
                    // if mBaseDir doesn't exist, files can be null.
                    if (files != null) {
                        for (int i = 0; i < files.length; i++) {
                            File f = new File(mBaseDir, files[i]);
                            if (!f.delete()) {
                                Log.e(LOGTAG, f.getPath() + " delete failed.");
                            }
                        }
                    }
                } catch (SecurityException e) {
                    // Ignore SecurityExceptions.
                }
            }
        };
        new Thread(clearCache).start();
        return true;
    }

    /**
     * Return true if the cache is empty.
     */
    static boolean cacheEmpty() {
        return mDataBase.hasCache();
    }

    static void trimCacheIfNeeded() {
        mDataBase.flushCacheStat();
        if (mDataBase.getCacheTotalSize() > CACHE_THRESHOLD) {
            List<String> pathList = mDataBase.trimCache(CACHE_TRIM_AMOUNT);
            int size = pathList.size();
            for (int i = 0; i < size; i++) {
                File f = new File(mBaseDir, pathList.get(i));
                if (!f.delete()) {
                    Log.e(LOGTAG, f.getPath() + " delete failed.");
                }
            }
            // remove the unreferenced files in the cache directory
            final List<String> fileList = mDataBase.getAllCacheFileNames();
            if (fileList == null) return;
            String[] toDelete = mBaseDir.list(new FilenameFilter() {
                public boolean accept(File dir, String filename) {
                    if (fileList.contains(filename)) {
                        return false;
                    } else {
                        return true;
                    }
                }
            });
            if (toDelete == null) return;
            size = toDelete.length;
            for (int i = 0; i < size; i++) {
                File f = new File(mBaseDir, toDelete[i]);
                if (!f.delete()) {
                    Log.e(LOGTAG, f.getPath() + " delete failed.");
                }
            }
        }
    }

    static void clearCache() {
        // delete database
        mDataBase.clearCache();
    }

    private static boolean checkCacheRedirect(int statusCode) {
        if (statusCode == 301 || statusCode == 302 || statusCode == 307) {
            // as 303 can't be cached, we do not return true
            return true;
        } else {
            return false;
        }
    }

    private static String getDatabaseKey(String url, long postIdentifier) {
        if (postIdentifier == 0) return url;
        return postIdentifier + url;
    }

    @SuppressWarnings("deprecation")
    private static void setupFiles(String url, CacheResult cacheRet) {
        if (true) {
            // Note: SHA1 is much stronger hash. But the cost of setupFiles() is
            // 3.2% cpu time for a fresh load of nytimes.com. While a simple
            // String.hashCode() is only 0.6%. If adding the collision resolving
            // to String.hashCode(), it makes the cpu time to be 1.6% for a 
            // fresh load, but 5.3% for the worst case where all the files 
            // already exist in the file system, but database is gone. So it
            // needs to resolve collision for every file at least once.
            int hashCode = url.hashCode();
            StringBuffer ret = new StringBuffer(8);
            appendAsHex(hashCode, ret);
            String path = ret.toString();
            File file = new File(mBaseDir, path);
            if (true) {
                boolean checkOldPath = true;
                // Check hash collision. If the hash file doesn't exist, just
                // continue. There is a chance that the old cache file is not
                // same as the hash file. As mDataBase.getCache() is more 
                // expansive than "leak" a file until clear cache, don't bother.
                // If the hash file exists, make sure that it is same as the 
                // cache file. If it is not, resolve the collision.
                while (file.exists()) {
                    if (checkOldPath) {
                        CacheResult oldResult = mDataBase.getCache(url);
                        if (oldResult != null && oldResult.contentLength > 0) {
                            if (path.equals(oldResult.localPath)) {
                                path = oldResult.localPath;
                            } else {
                                path = oldResult.localPath;
                                file = new File(mBaseDir, path);
                            }
                            break;
                        }
                        checkOldPath = false;
                    }
                    ret = new StringBuffer(8);
                    appendAsHex(++hashCode, ret);
                    path = ret.toString();
                    file = new File(mBaseDir, path);
                }
            }
            cacheRet.localPath = path;
            cacheRet.outFile = file;
        } else {
            // get hash in byte[]
            Digest digest = new SHA1Digest();
            int digestLen = digest.getDigestSize();
            byte[] hash = new byte[digestLen];
            int urlLen = url.length();
            byte[] data = new byte[urlLen];
            url.getBytes(0, urlLen, data, 0);
            digest.update(data, 0, urlLen);
            digest.doFinal(hash, 0);
            // convert byte[] to hex String
            StringBuffer result = new StringBuffer(2 * digestLen);
            for (int i = 0; i < digestLen; i = i + 4) {
                int h = (0x00ff & hash[i]) << 24 | (0x00ff & hash[i + 1]) << 16
                        | (0x00ff & hash[i + 2]) << 8 | (0x00ff & hash[i + 3]);
                appendAsHex(h, result);
            }
            cacheRet.localPath = result.toString();
            cacheRet.outFile = new File(mBaseDir, cacheRet.localPath);
        }
    }

    private static void appendAsHex(int i, StringBuffer ret) {
        String hex = Integer.toHexString(i);
        switch (hex.length()) {
            case 1:
                ret.append("0000000");
                break;
            case 2:
                ret.append("000000");
                break;
            case 3:
                ret.append("00000");
                break;
            case 4:
                ret.append("0000");
                break;
            case 5:
                ret.append("000");
                break;
            case 6:
                ret.append("00");
                break;
            case 7:
                ret.append("0");
                break;
        }
        ret.append(hex);
    }

    private static CacheResult parseHeaders(int statusCode, Headers headers,
            String mimeType) {
        // if the contentLength is already larger than CACHE_MAX_SIZE, skip it
        if (headers.getContentLength() > CACHE_MAX_SIZE) return null;

        // The HTML 5 spec, section 6.9.4, step 7.3 of the application cache
        // process states that HTTP caching rules are ignored for the
        // purposes of the application cache download process.
        // At this point we can't tell that if a file is part of this process,
        // except for the manifest, which has its own mimeType.
        // TODO: work out a way to distinguish all responses that are part of
        // the application download process and skip them.
        if (MANIFEST_MIME.equals(mimeType)) return null;

        // TODO: if authenticated or secure, return null
        CacheResult ret = new CacheResult();
        ret.httpStatusCode = statusCode;

        String location = headers.getLocation();
        if (location != null) ret.location = location;

        ret.expires = -1;
        ret.expiresString = headers.getExpires();
        if (ret.expiresString != null) {
            try {
                ret.expires = AndroidHttpClient.parseDate(ret.expiresString);
            } catch (IllegalArgumentException ex) {
                // Take care of the special "-1" and "0" cases
                if ("-1".equals(ret.expiresString)
                        || "0".equals(ret.expiresString)) {
                    // make it expired, but can be used for history navigation
                    ret.expires = 0;
                } else {
                    Log.e(LOGTAG, "illegal expires: " + ret.expiresString);
                }
            }
        }

        String contentDisposition = headers.getContentDisposition();
        if (contentDisposition != null) {
            ret.contentdisposition = contentDisposition;
        }

        String crossDomain = headers.getXPermittedCrossDomainPolicies();
        if (crossDomain != null) {
            ret.crossDomain = crossDomain;
        }

        // lastModified and etag may be set back to http header. So they can't
        // be empty string.
        String lastModified = headers.getLastModified();
        if (lastModified != null && lastModified.length() > 0) {
            ret.lastModified = lastModified;
        }

        String etag = headers.getEtag();
        if (etag != null && etag.length() > 0) ret.etag = etag;

        if (speculativeEnabled()) {
        	Log.d("speculative", "CacheManager: parseHeaders");
        }
        
        String cacheControl = headers.getCacheControl();
        if (cacheControl != null) {
            String[] controls = cacheControl.toLowerCase().split("[ ,;]");
            boolean noCache = false;
            for (int i = 0; i < controls.length; i++) {
                if (NO_STORE.equals(controls[i])) {
                    if (!speculativeEnabled()) {
                        return null;
                    }
                }
                // According to the spec, 'no-cache' means that the content
                // must be re-validated on every load. It does not mean that
                // the content can not be cached. set to expire 0 means it
                // can only be used in CACHE_MODE_CACHE_ONLY case
                if (NO_CACHE.equals(controls[i])) {
                    ret.expires = 0;
                    noCache = true;
                // if cache control = no-cache has been received, ignore max-age
                // header, according to http spec:
                // If a request includes the no-cache directive, it SHOULD NOT
                // include min-fresh, max-stale, or max-age.
                } else if (controls[i].startsWith(MAX_AGE) && !noCache) {
                    int separator = controls[i].indexOf('=');
                    if (separator < 0) {
                        separator = controls[i].indexOf(':');
                    }
                    if (separator > 0) {
                        String s = controls[i].substring(separator + 1);
                        try {
                            long sec = Long.parseLong(s);
                            if (sec >= 0) {
                                ret.expires = System.currentTimeMillis() + 1000
                                        * sec;
                            }
                        } catch (NumberFormatException ex) {
                            if ("1d".equals(s)) {
                                // Take care of the special "1d" case
                                ret.expires = System.currentTimeMillis() + 86400000; // 24*60*60*1000
                            } else {
                                Log.e(LOGTAG, "exception in parseHeaders for "
                                        + "max-age:"
                                        + controls[i].substring(separator + 1));
                                ret.expires = 0;
                            }
                        }
                    }
                }
            }
        }

        // According to RFC 2616 section 14.32:
        // HTTP/1.1 caches SHOULD treat "Pragma: no-cache" as if the
        // client had sent "Cache-Control: no-cache"
        if (NO_CACHE.equals(headers.getPragma())) {
            ret.expires = 0;
        }

        // According to RFC 2616 section 13.2.4, if an expiration has not been
        // explicitly defined a heuristic to set an expiration may be used.
        if (ret.expires == -1) {
            if (ret.httpStatusCode == 301) {
                // If it is a permanent redirect, and it did not have an
                // explicit cache directive, then it never expires
                ret.expires = Long.MAX_VALUE;
            } else if (ret.httpStatusCode == 302 || ret.httpStatusCode == 307) {
                // If it is temporary redirect, expires
                ret.expires = 0;
            } else if (ret.lastModified == null) {
                // When we have no last-modified, then expire the content with
                // in 24hrs as, according to the RFC, longer time requires a
                // warning 113 to be added to the response.

                // Only add the default expiration for non-html markup. Some
                // sites like news.google.com have no cache directives.
                if (!mimeType.startsWith("text/html")) {
                    ret.expires = System.currentTimeMillis() + 86400000; // 24*60*60*1000
                } else {
                    // Setting a expires as zero will cache the result for
                    // forward/back nav.
                    ret.expires = 0;
                }
            } else {
                // If we have a last-modified value, we could use it to set the
                // expiration. Suggestion from RFC is 10% of time since
                // last-modified. As we are on mobile, loads are expensive,
                // increasing this to 20%.

                // 24 * 60 * 60 * 1000
                long lastmod = System.currentTimeMillis() + 86400000;
                try {
                    lastmod = AndroidHttpClient.parseDate(ret.lastModified);
                } catch (IllegalArgumentException ex) {
                    Log.e(LOGTAG, "illegal lastModified: " + ret.lastModified);
                }
                long difference = System.currentTimeMillis() - lastmod;
                if (difference > 0) {
                    ret.expires = System.currentTimeMillis() + difference / 5;
                } else {
                    // last modified is in the future, expire the content
                    // on the last modified
                    ret.expires = lastmod;
                }
            }
        }

        return ret;
    }
    
    private static boolean mSpeculativeEnabled = false;	// for speculative loading; init as disabled
    private static boolean mCacheExpireEnabled = false;	// for checking cache expiration; true = always expire
    
    private static ArrayList<String> mPredictedQueue; // predicted subresources
    private static ArrayList<FrameLoader> mPredictedWaitQueue; // predicted but not speculatively loaded resources
    private static ArrayList<String> mSpeculativeLoadingPool; // predicted and is speculatively loaded resources
    private static ArrayList<String> mSpeculativeFinishedPool; // speculatively loaded resources that finished
    private static ArrayList<String> mSpeculativeRedirectPool; // speculatively loaded resources that are redirected
    private static ArrayList<LoadListener> mSpeculativeErrorPool; // speculatively loaded resources that have erros
    private static ArrayList<String> mNoStorePool; // resources that should not be stored
    private static ArrayList<FrameLoader> mNormalLoaderPool; // normal frame loaders that has been speculatively loaded before by speculative loader
        
    private static void speculativeInit() {
        //mSpeculativeEnabled = false; // this will be set/cleared in setting when opening browser
        mCacheExpireEnabled = false;
        mPredictedQueue = new ArrayList<String>(20);
        mPredictedWaitQueue = new ArrayList<FrameLoader>(20);
        mSpeculativeLoadingPool = new ArrayList<String>(20);
        mSpeculativeFinishedPool = new ArrayList<String>(20);
        mSpeculativeRedirectPool = new ArrayList<String>(20);
        mSpeculativeErrorPool = new ArrayList<LoadListener>(20);
        mNormalLoaderPool = new ArrayList<FrameLoader>(20);
        mNoStorePool = new ArrayList<String>(20);
        Log.d("speculative", "CacheManager: speculativeInit()");
    }
    /**
     * @hide
     */
    public static void setSpeculativeEnabled() {
       	// load resource metadata when enabling it
		if (CacheManager.isDirtyResourceMetadata()) {
			// init to dirty, so will load on start
			CacheManager.loadResourceMetadata();
			// speculation will be enabled in loadThread
		}
		else {
			mSpeculativeEnabled = true;
			Log.d("speculative", "\n*********************************************\n"
								+"*********** speculation enabled *************\n"
								+"*********************************************\n\n");
			Log.d("speculative", "clean resource metadata, no need to load");
		}
	}
    /**
     * @hide
     */
    public static void clearSpeculativeEnabled() {
       	Log.d("speculative", "\n*********************************************\n"
       						+"*********** speculation disabled ************\n"
       						+"*********************************************\n\n");
   		mSpeculativeEnabled = false;
    }
    /**
     * @hide
     */
    public static boolean speculativeEnabled() {
        //Log.d("speculative", "CacheManager: speculativeEnabled()");
    	return mSpeculativeEnabled;
    }
    /**
     * @hide
     */
    public static boolean cacheExpireEnabled() {
        //Log.d("speculative", "CacheManager: cacheExpireEnabled()");
    	return mCacheExpireEnabled;
    }
    /**
     * @hide
     */
    public static void speculativeFinish() {
     	 clearPredictedQueue();
     	 clearPredictedWaitQueue();
     	 clearSpeculativeLoadingPool();
     	 clearSpeculativeFinishedPool();
     	 clearSpeculativeRedirectPool();
     	 clearSpeculativeErrorPool();
     	 clearNormalLoaderPool();
     	 clearNoStorePool();
     	 Log.d("speculative", "CacheManager: speculativeFinish()");
     }
    /**
     * @hide
     */
    public synchronized static void addToPredictedQueue(String url) {
        //Log.d("speculative", "CacheManager: addToPredictedQueue() | url="+url);
    	mPredictedQueue.add(url);
    }
    /**
     * @hide
     */
    public synchronized static void addToPredictedWaitQueue(FrameLoader loader) {
        //Log.d("speculative", "CacheManager: addToPredictedWaitQueue() | url="+url);
    	mPredictedWaitQueue.add(loader);
    }
    /**
     * @hide
     */
    public synchronized static void addToSpeculativeLoadingPool(String url) {
        //Log.d("speculative", "CacheManager: addToSpeculativeLoadingPool() | url="+url);
    	mSpeculativeLoadingPool.add(url);
    }
    /**
     * @hide
     */
    public synchronized static void addToSpeculativeFinishedPool(String url) {
        //Log.d("speculative", "CacheManager: addToSpeculativeFinishedPool() | url="+url);
    	mSpeculativeFinishedPool.add(url);
    }
    /**
     * @hide
     */
    public synchronized static void addToSpeculativeRedirectPool(String url) {
        //Log.d("speculative", "CacheManager: addToSpeculativeRedirectPool() | url="+url);
    	mSpeculativeRedirectPool.add(url);
    }
    /**
     * @hide
     */
    public synchronized static void addToSpeculativeErrorPool(LoadListener listener) {
        //Log.d("speculative", "CacheManager: addToSpeculativeErrorPool() | url="+url);
    	mSpeculativeErrorPool.add(listener);
    }
    /**
     * @hide
     */
    public synchronized static void addToNormalLoaderPool(FrameLoader loader) {
        //Log.d("speculative", "CacheManager: addToNormalLoaderPool() | url="+loader.getLoadListener().url());
    	mNormalLoaderPool.add(loader);
    }
    /**
     * @hide
     */
    public synchronized static void addToNoStorePool(String url) {
        //Log.d("speculative", "CacheManager: addToNoStorePool() | url="+url);
    	mNoStorePool.add(url);
    }
    /**
     * @hide
     */
    public synchronized static void removeFromPredictedWaitQueue(String url) {
        //Log.d("speculative", "CacheManager: removeFromPredictedWaitQueue() | url="+url);
        // notice: cannot use contains() here, it compares reference
        for (int i=0; i<mPredictedWaitQueue.size(); i++){
        	if (url.equals(mPredictedWaitQueue.get(i).getLoadListener().url())) {
        		mPredictedWaitQueue.remove(i);
        	}
        }
    }
    /**
     * @hide
     */
    public synchronized static FrameLoader popFromPredictedWaitQueue() {
        //Log.d("speculative", "CacheManager: popFromPredictedWaitQueue() | url="+url);
        if (mPredictedWaitQueue.isEmpty()) {
        	return null;
        }
        else {
        	return mPredictedWaitQueue.remove(0);
        }
    }
    /**
     * @hide
     */
    public synchronized static void removeFromSpeculativeLoadingPool(String url) {
        //Log.d("speculative", "CacheManager: removeFromSpeculativeLoadingPool() | url="+url);
        // notice: cannot use contains() here, it compares reference
        for (int i=0; i<mSpeculativeLoadingPool.size(); i++){
        	if (url.equals(mSpeculativeLoadingPool.get(i))) {
        		mSpeculativeLoadingPool.remove(i);
        	}
        }
    }
    /**
     * @hide
     */
    public synchronized static LoadListener getSpeculativeErrorLoader(String url) {
    	LoadListener listener = null;
    	for (int i = 0; i < mSpeculativeErrorPool.size(); i++) {
    		listener = mSpeculativeErrorPool.get(i);
    		if (listener != null) {
    			if (url.equals(listener.url())) {
    				return listener; // don't remove now
    			}
    		}
    	}
    	return null;
    }
    /**
     * @hide
     */
    public synchronized static FrameLoader getWaitingNormalLoader(String url) {
    	FrameLoader loader = null;
    	for (int i = 0; i < mNormalLoaderPool.size(); i++) {
    		loader = mNormalLoaderPool.get(i);
    		if (loader != null) {
    			if (url.equals(loader.getLoadListener().url())) {
    				Log.d("speculative", "CacheManager: getWaitingNormalLoader - found! | url="+url);
    				return mNormalLoaderPool.remove(i);
    			}
    		}
    	}
    	Log.d("speculative", "CacheManager: getWaitingNormalLoader (not waiting) | url="+url);
    	return null;
    }
    /**
     * @hide
     */
    public synchronized static boolean inPredictedQueue(String url) {
        //Log.d("speculative", "CacheManager: inPredictedQueue() | url="+url);
        // notice: cannot use contains() here, it compares reference
        for (int i=0; i<mPredictedQueue.size(); i++){
        	if (url.equals(mPredictedQueue.get(i))) {
        		return true;
        	}
        }
        return false;
    }
    /**
     * @hide
     */
    public synchronized static boolean inPredictedWaitQueue(String url) {
        //Log.d("speculative", "CacheManager: inPredictedWaitQueue() | url="+url);
        // notice: cannot use contains() here, it compares reference
        for (int i=0; i<mPredictedWaitQueue.size(); i++){
        	if (url.equals(mPredictedWaitQueue.get(i).getLoadListener().url())) {
        		return true;
        	}
        }
        return false;
    }
    /**
     * @hide
     */
    public synchronized static boolean inSpeculativeLoadingPool(String url) {
        //Log.d("speculative", "CacheManager: inSpeculativeLoadingPool() | url="+url);
        // notice: cannot use contains() here, it compares reference
        for (int i=0; i<mSpeculativeLoadingPool.size(); i++) {
        	if(url.equals(mSpeculativeLoadingPool.get(i))) {
        		return true;
        	}
        }
        return false;
    }
    /**
     * @hide
     */
    public synchronized static boolean inSpeculativeFinishedPool(String url) {
        //Log.d("speculative", "CacheManager: inSpeculativeFinishedPool() | url="+url);
        // notice: cannot use contains() here, it compares reference
        for (int i=0; i<mSpeculativeFinishedPool.size(); i++) {
        	if(url.equals(mSpeculativeFinishedPool.get(i))) {
        		return true;
        	}
        }
        return false;
    }
    /**
     * @hide
     */
    public synchronized static boolean inSpeculativeRedirectPool(String url) {
        //Log.d("speculative", "CacheManager: inSpeculativeRedirectPool() | url="+url);
        // notice: cannot use contains() here, it compares reference
        for (int i=0; i<mSpeculativeRedirectPool.size(); i++) {
        	if(url.equals(mSpeculativeRedirectPool.get(i))) {
        		return true;
        	}
        }
        return false;
    }
    /**
     * @hide
     */
    public synchronized static boolean inSpeculativeErrorPool(String url) {
        //Log.d("speculative", "CacheManager: inSpeculativeErrorPool() | url="+url);
    	LoadListener listener = null;
    	for (int i = 0; i < mSpeculativeErrorPool.size(); i++) {
    		listener = mSpeculativeErrorPool.get(i);
    		if (listener != null) {
    			if (url.equals(listener.url())) {
    				return true;
    			}
    		}
    	}
    	return false;
    }
    
    private synchronized static void clearPredictedQueue() {
    	mPredictedQueue.clear();
        Log.d("speculative", "CacheManager: clearPredictedQueue DONE");
    }
    public synchronized static void clearPredictedWaitQueue() {
    	// get called after main resource tearDown
    	int size = mPredictedWaitQueue.size();
    	mPredictedWaitQueue.clear();
        Log.d("speculative", "CacheManager: clearPredictedWaitQueue ("
        	+Integer.toString(size)+" predicted subres not loaded speculatively)");
    }
    private synchronized static void clearSpeculativeLoadingPool() {
    	mSpeculativeLoadingPool.clear();
        Log.d("speculative", "CacheManager: clearSpeculativeLoadingPool DONE");
    }
    private synchronized static void clearSpeculativeFinishedPool() {
    	mSpeculativeFinishedPool.clear();
        Log.d("speculative", "CacheManager: clearSpeculativeFinishedPool DONE");
    }
    private synchronized static void clearSpeculativeRedirectPool() {
    	mSpeculativeRedirectPool.clear();
        Log.d("speculative", "CacheManager: clearSpeculativeRedirectPool DONE");
    }
    private synchronized static void clearSpeculativeErrorPool() {
    	mSpeculativeErrorPool.clear();
        Log.d("speculative", "CacheManager: clearSpeculativeErrorPool DONE");
    }
    private synchronized static void clearNormalLoaderPool() {
   		FrameLoader loader = null;
   		for (int i=0; i<mNormalLoaderPool.size(); i++) {
   			loader = mNormalLoaderPool.get(i);
   			loader.executeLoad();
   			Log.d("speculative", "CacheManager: clearNormalLoaderPool | url="+loader.getLoadListener().url());
   		}
    	mNormalLoaderPool.clear();
        Log.d("speculative", "CacheManager: clearNormalLoaderPool DONE");
    }
    private synchronized static void clearNoStorePool() {
    	int size = mNoStorePool.size();
    	for (int i = 0; i < size; i++) {
    		CacheResult result = mDataBase.getCache(mNoStorePool.get(i));
    		if (result != null) {
    			File f = new File(mBaseDir, result.localPath);
                if (!f.delete()) {
                    Log.e(LOGTAG, f.getPath() + " delete failed.");
                }
                mDataBase.removeCache(mNoStorePool.get(i));
    		}
    	}
    	mNoStorePool.clear();
    	Log.d("speculative", "CacheManager: clearNoStorePool DONE");
    }
    

    
    /**
     * @hide
     */
    public static class ResourceNode implements Serializable{
    	String type;
    	String url;
    	long lvt;  		// last visit time in epoch milliseconds
    	long nov; 		// number of visits, for subres nodes only
    	int priority; 	// for subres nodes only: JS(1); CSS(2); others(3);
    	ArrayList<ResourceNode> children;	// for website, subdomain and webpage nodes
    	ArrayList<ResourceNode> parents;	// for subres nodes
    }
    
    /**
     * @hide
     */
    public static class ResourceGraph implements Serializable{
    	private long mNumOfVisits; // number of visits for the site
    	private ResourceNode mRoot;
    	private ArrayList<ResourceNode> mSubresNodes;
    	
    	public long getNumOfVisits() {
    		return mNumOfVisits;
    	}
    	
    	public String getSite() {
    		return mRoot.url;
    	}
    	
    	ResourceGraph(String site, long currentTime) {
    		mNumOfVisits = 0;
    		mRoot = createNode("website", site, currentTime);
    		mSubresNodes = new ArrayList<ResourceNode>();
    	}
    	
    	private ResourceNode createNode(String nodeType, String url, long currentTime) {
    		ResourceNode node = new ResourceNode();
    		node.type = nodeType;
    		node.url = url;
    		node.lvt = currentTime;
    		node.nov = 0;
    		node.priority = 0;
    		if (nodeType.equals("subres")) {
    			if (URLUtil.isJS(url)) {
    				node.priority = 1;
    			}
    			else if (URLUtil.isCSS(url)) {
    				node.priority = 2;
    			}
    			else {
    				node.priority = 3;
    			}
    			node.children = null;
    			node.parents = new ArrayList<ResourceNode>();
    		}
    		else if (nodeType.equals("webpage")||nodeType.equals("subdomain")||nodeType.equals("website")) {
    			node.children = new ArrayList<ResourceNode>();
    			node.parents = null;
    		}
    		else {
    			node.children = null;
    			node.parents = null;
    		}
    		return node;
    	}
    	
    	private ResourceNode getSubdomainNode(String subdomain) {
    		ResourceNode subdomain_node = null;
    		for (int i=0; i<mRoot.children.size(); i++) {
    			subdomain_node = mRoot.children.get(i);
    			if (subdomain_node.url.equals(subdomain)) {
    				return subdomain_node;
    			}
    		}
    		return null;
    	}
    	
    	private ResourceNode getWebpageNode(String webpage) {
    		String subdomain = URLUtil.getNetloc(webpage);
    		ResourceNode subdomain_node = null;
    		ResourceNode webpage_node = null;
    		for (int i=0; i<mRoot.children.size(); i++) {
    			subdomain_node = mRoot.children.get(i);
    			if (subdomain_node.url.equals(subdomain)) {
    				for (int j=0; j<subdomain_node.children.size(); j++) {
    					webpage_node = subdomain_node.children.get(j);
    					if (webpage_node.url.equals(webpage)) {
    						return webpage_node;
    					}
    				}
    			}
    		}
    		return null;
    	}
    	
    	private ResourceNode getSubresNode(String subres) {
    		ResourceNode subres_node = null;
    		for (int i=0; i<mSubresNodes.size(); i++) {
    			subres_node = mSubresNodes.get(i);
    			if (subres_node.url.equals(subres)) {
    				return subres_node;
    			}
    		}
    		return null;
    	}
    	
    	private int avgWebpageOutDegree(ResourceNode node) {
    		int total_num_subres = 0;
    		int num_webpages_counted = 0;
    		if (node.type.equals("website")) {
    			ResourceNode subdomain_node = null;
    			ResourceNode webpage_node = null;
    			for (int i=0; i<mRoot.children.size(); i++) {
    				subdomain_node = mRoot.children.get(i);
    				for (int j=0; j<subdomain_node.children.size(); j++) {
    					webpage_node = subdomain_node.children.get(j);
    					total_num_subres += webpage_node.children.size();
    					num_webpages_counted += 1;
    				}
    			}
    		}
    		else if (node.type.equals("subdomain")) {
    			ResourceNode webpage_node = null;
    			for (int j=0; j<node.children.size(); j++) {
    				webpage_node = node.children.get(j);
    				total_num_subres += webpage_node.children.size();
    				num_webpages_counted += 1;
    			}
    		}
    		
    		Log.d("speculative", "avgWebpageOutDegree: "
    			+"total_num_subres ("+Integer.toString(total_num_subres)
    			+") num_webpages_counted ("+Integer.toString(num_webpages_counted)+")");
    		if (num_webpages_counted == 0) {
    			return 0;
    		}
    		else {
    			Log.d("speculative", "avgWebpageOutDegree: return "+Integer.toString((int)(total_num_subres/num_webpages_counted)));
    			return (int)(total_num_subres/num_webpages_counted);
    		}
    	}
    	
    	// update resource graph
    	public void update(String webpage, String[] subresources, long currentTime) {
    		// update total number of visits of the site
    		mNumOfVisits += 1;
    		// update website root node
    		mRoot.lvt = currentTime;
    		// update/add subdomain node
    		String subdomain = URLUtil.getNetloc(webpage);
    		ResourceNode subdomain_node = getSubdomainNode(subdomain);
    		if (subdomain_node == null) {
    			subdomain_node = createNode("subdomain", subdomain, currentTime);
    			mRoot.children.add(subdomain_node);
    		}
    		else {
    			subdomain_node.lvt = currentTime;
    		}
    		// update/add webpage node
    		ResourceNode webpage_node = getWebpageNode(webpage);
    		if (webpage_node == null) {
    			webpage_node = createNode("webpage", webpage, currentTime);
    			subdomain_node.children.add(webpage_node);
    		}
    		else {
    			webpage_node.lvt = currentTime;
    		}
    		// update/add subres nodes
    		ArrayList<String> fresh_subres_list = new ArrayList<String>(10);
    		ResourceNode subres_node = null;
    		for (int i=webpage_node.children.size()-1; i>=0; i--) {
    			subres_node = webpage_node.children.get(i);
    			// update un-stale (fresh) subres nodes
    			boolean subres_stale = true;
    			for (int l=0; l<subresources.length; l++) {
    				if (subres_node.url.equals(subresources[l])) {
    					subres_node.lvt = currentTime;
    					subres_node.nov += 1;
    					fresh_subres_list.add(subresources[l]);
    					subres_stale = false;
    					break;
    				}
    			}
    			// remove stale subrs
    			if (subres_stale) {
    				// remove webpage from subres node's parent list
    				for (int j=0; j<subres_node.parents.size(); j++) {
    					if (subres_node.parents.get(j).url.equals(webpage)) {
    						subres_node.parents.remove(j);
    						break;
    					}
    				}
    				// remove subres from webpage node's children list
    				webpage_node.children.remove(i);
    				// remove the subres node from mSubresNodes if no webpage uses it
    				if (subres_node.parents.size() == 0) {
    					for (int k=0; k<mSubresNodes.size(); k++) {
    						if (subres_node.url.equals(mSubresNodes.get(k).url)) {
    							mSubresNodes.remove(k);
    							break;
    						}
    					}
    				}
    			}
    		}
    		// add new subresources to webpage node; create subres node if not exist
    		for (int i=0; i<subresources.length; i++) {
    			boolean isNewSubres = true;
    			for (int j=0; j<fresh_subres_list.size(); j++) {
    				if (subresources[i].equals(fresh_subres_list.get(j))) {
    					isNewSubres = false;
    					break;
    				}
    			}
    			if (isNewSubres) {
    				subres_node = getSubresNode(subresources[i]);
    				// subres node exist, update it
    				if (subres_node != null) {
    					subres_node.lvt = currentTime;
    				}
    				// subres node not exist, create one
    				else {
    					subres_node = createNode("subres", subresources[i], currentTime);
    					mSubresNodes.add(subres_node);
    				}
    				subres_node.nov += 1;
    				subres_node.parents.add(webpage_node);
    				webpage_node.children.add(subres_node);
    			}
    		}
    	}
    	
    	// trim resource graph by removing outdated nodes
    	// last visited time: 30 days ago (2592000000 milliseconds)
    	public void trim(long currentTime) {
    		long threshold = 2592000000L;
    		ResourceNode subdomain_node = null;
    		ResourceNode webpage_node = null;
    		ResourceNode subres_node = null;
    		// go through the graph
    		for (int i=mRoot.children.size()-1; i>=0; i--) {
    			subdomain_node = mRoot.children.get(i);
    			// remove whole subdomain
    			if (subdomain_node.lvt + threshold < currentTime) {
    				for (int j=subdomain_node.children.size()-1; j>=0; j--) {
    					webpage_node = subdomain_node.children.get(j);
    					for (int k=webpage_node.children.size()-1; k>=0; k--) {
    						subres_node = webpage_node.children.get(k);
    						for (int l=subres_node.parents.size()-1; l>=0; l--) {
    							if (subres_node.parents.get(l).url.equals(webpage_node.url)) {
    								subres_node.parents.remove(l);
    							}
    						}
    					}
    					webpage_node.children.clear();
    				}
    				subdomain_node.children.clear();
    				mRoot.children.remove(i);
    			}
    			else {
    				for (int j=subdomain_node.children.size()-1; j>=0; j--) {
    					webpage_node = subdomain_node.children.get(j);
    					// remove webpage node and its children
    					if (webpage_node.lvt + threshold < currentTime) {
							for (int k=webpage_node.children.size()-1; k>=0; k--) {
								subres_node = webpage_node.children.get(k);
								for (int l=subres_node.parents.size()-1; l>=0; l--) {
									if (subres_node.parents.get(l).url.equals(webpage_node.url)) {
										subres_node.parents.remove(l);
									}
								}
							}
							webpage_node.children.clear();
							subdomain_node.children.remove(j);
    					}
    					else {
    						// remove stale subres nodes in a fresh webpage:
    						// should not happen because subres nodes are updated
    						// when updating webpage nodes, fresh subres nodes
    						// have the same lvt as the webpage node
    					}
    				}
    			}
    		}
    		// remove stale subres nodes from mSubresNodes
    		for (int i=mSubresNodes.size()-1; i>=0; i--) {
    			subres_node = mSubresNodes.get(i);
    			if (subres_node.lvt+threshold<currentTime || subres_node.parents.size() == 0) {
    				subres_node.parents.clear();
    				mSubresNodes.remove(i);
    			}
    		}
    	}
    	
    	// predict the subresources for webpage
    	public String[] predict(String webpage) {
    		ResourceNode webpage_node = getWebpageNode(webpage);
    		// webpage revisit
    		if (webpage_node != null) {
    			Log.d("speculative", "predict: revisit!");
    			ArrayList<ResourceNode> target_subres_nodes = webpage_node.children;
    			if (target_subres_nodes.size() == 0) {
    				return null;
    			}
    			else {
					// sort by file priority, number of visits and url length
					ResourceNode tmp = null;
					for (int i=0; i<target_subres_nodes.size()-1; i++) {
						for (int j=0; j<target_subres_nodes.size()-(i+1); j++) {
							// sort by file priority: JS(1) -> CSS(2) -> image(3)
							int priority_j = target_subres_nodes.get(j).priority;
							int priority_j_1 = target_subres_nodes.get(j+1).priority;
							if (priority_j > priority_j_1) {
								Collections.swap(target_subres_nodes, j, j+1);
							}
							else if (priority_j == priority_j_1) {
								// then sort by number of visits: large -> small
								long nov_j = target_subres_nodes.get(j).nov;
								long nov_j_1 = target_subres_nodes.get(j+1).nov;
								if (nov_j < nov_j_1) {
									Collections.swap(target_subres_nodes, j, j+1);
								}
								else if (nov_j == nov_j_1) {
									// finally by url length: short -> long
									int length_j = target_subres_nodes.get(j).url.length();
									int length_j_1 = target_subres_nodes.get(j+1).url.length();
									if (length_j > length_j_1) {
										Collections.swap(target_subres_nodes, j, j+1);
									}
								}
							}
						}
					}
					String[] predicted_subres_nodes = new String[target_subres_nodes.size()];
					for (int i=0; i<target_subres_nodes.size(); i++) {
						predicted_subres_nodes[i] = target_subres_nodes.get(i).url;
						Log.d("speculative", " - "+predicted_subres_nodes[i]);
					}
					return predicted_subres_nodes;
    			}
    		}
    		// webpage new visit
    		else {
    			Log.d("speculative", "predict: new visit!");
    			// get num_predicted and target_subres_nodes
    			int num_predicted = 0;
    			ArrayList<ResourceNode> target_subres_nodes = new ArrayList<ResourceNode>();
    			ResourceNode subdomain_node = getSubdomainNode(URLUtil.getNetloc(webpage));
    			if (subdomain_node != null) {
    				num_predicted = avgWebpageOutDegree(subdomain_node);
    				webpage_node = null;
    				ResourceNode subres_node = null;
    				for (int i=0; i<subdomain_node.children.size(); i++) {
    					webpage_node = subdomain_node.children.get(i);
    					for (int j=0; j<webpage_node.children.size(); j++) {
    						subres_node = webpage_node.children.get(j);
    						if (!target_subres_nodes.contains(subres_node)) {
    							target_subres_nodes.add(subres_node);
    						}
    					}
    				}
    			}
    			else {
    				num_predicted = avgWebpageOutDegree(mRoot);
    				target_subres_nodes = mSubresNodes;
    			}
    			// get the predicted subresources
    			if (num_predicted==0 || target_subres_nodes.size()==0) {
    				return null;
    			}
    			else {
					// sort target_subres_nodes by ref_count, file type, number of visit and url length
					// longer url may be session dependent or have redirect
					ResourceNode tmp = null;
					for (int i=0; i<target_subres_nodes.size()-1; i++) {
						for (int j=0; j<target_subres_nodes.size()-(i+1); j++) {
							// sort by ref_count: large -> small
							int size_j = target_subres_nodes.get(j).parents.size();
							int size_j_1 = target_subres_nodes.get(j+1).parents.size();
							if (size_j < size_j_1) {
								Collections.swap(target_subres_nodes, j, j+1);
							}
							else if (size_j == size_j_1) {
								// then sort by file priority: JS(1) -> CSS(2) -> image(3)
								int priority_j = target_subres_nodes.get(j).priority;
								int priority_j_1 = target_subres_nodes.get(j+1).priority;
								if (priority_j > priority_j_1) {
									Collections.swap(target_subres_nodes, j, j+1);
								}
								else if (priority_j == priority_j_1) {
									// then sort by number of visits: large -> small
									long nov_j = target_subres_nodes.get(j).nov;
									long nov_j_1 = target_subres_nodes.get(j+1).nov;
									if (nov_j < nov_j_1) {
										Collections.swap(target_subres_nodes, j, j+1);
									}
									else if (nov_j == nov_j_1) {
										// finally by url length: short -> long
										int length_j = target_subres_nodes.get(j).url.length();
										int length_j_1 = target_subres_nodes.get(j+1).url.length();
										if (length_j > length_j_1) {
											Collections.swap(target_subres_nodes, j, j+1);
										}
									}
								}
							}
						}
					}
					// choose the num_predicted subres nodes as the predicted subres nodes
					if (num_predicted > target_subres_nodes.size()) {
						num_predicted = target_subres_nodes.size();
					}
					String[] predicted_subres_nodes = new String[num_predicted];
					for (int i=0; i<num_predicted; i++) {
						predicted_subres_nodes[i] = target_subres_nodes.get(i).url;
						Log.d("speculative", " - "+predicted_subres_nodes[i]);
					}
					return predicted_subres_nodes;
				}
    		}
    	}
    	
    	// dump the resource graph
    	public void dump() {
    		ResourceNode subdomain_node = null;
    		ResourceNode webpage_node = null;
    		ResourceNode subres_node = null;
    		Log.d("speculative", " - "+mRoot.url);
    		for (int i=0; i<mRoot.children.size(); i++) {
    			subdomain_node = mRoot.children.get(i);
    			Log.d("speculative", "   - "+subdomain_node.url);
    			for (int j=0; j<subdomain_node.children.size(); j++) {
    				webpage_node = subdomain_node.children.get(j);
    				Log.d("speculative", "     - "+webpage_node.url);
    				for (int k=0; k<webpage_node.children.size(); k++) {
    					subres_node = webpage_node.children.get(k);
    					Log.d("speculative", "       - "+subres_node.url);
    				}
    			}
    		}
    	}
    }
    
    private static String mCurrentWebpage = null;
    private static ArrayList<String> mCurrentSubresources = new ArrayList<String>();
    /**
     * @hide
     */
    public static String getCurrentWebpage() {
    	return mCurrentWebpage;
    }
    /**
     * @hide
     */
    public static void setCurrentWebpage(String webpage) {
    	mCurrentWebpage = webpage;
    }
    /**
     * @hide
     */
    public static void addCurrentSubresource(String subresource) {
    	// only save network urls
    	if (URLUtil.isNetworkUrl(subresource)) {
    		for (int i=0; i<mCurrentSubresources.size(); i++) {
    			// no need to add if already added
    			if (subresource.equals(mCurrentSubresources.get(i))) {
    				return;
    			}
    		}
    		mCurrentSubresources.add(subresource);
    	}
    }
    /**
     * @hide
     */
    private static void clearCurrentWebpageSubresources() {
    	mCurrentWebpage = null;
    	if (mCurrentSubresources!=null) {
    		mCurrentSubresources.clear();
    	}
    }
    
    private static boolean mDirtyResourceMetadata = true; // init to dirty so will load at the beginning
    private static ArrayList<ResourceGraph> mResourceMetadata = null;
    private static String mResourceMetadataFileName = "/sdcard/speculative/ResourceMetadata.ser";
    
    /**
     * @hide
     */
    public static boolean isDirtyResourceMetadata() {
    	return mDirtyResourceMetadata;
    }
    /**
     * @hide
     */
    private static void dumpResourceMetadata() {
    	if (mResourceMetadata != null) {
			ResourceGraph resource_graph = null;
			for (int i=0; i<mResourceMetadata.size(); i++) {
				resource_graph = mResourceMetadata.get(i);
				resource_graph.dump();
			}
    	}
    	
        Log.d("speculative", "CACHE_THRESHOLD: "
        	+Long.toString(SystemProperties.getLong("nw.cache.threshold", CACHE_THRESHOLD_DEF)));
        Log.d("speculative", "CACHE_TRIM_AMOUNT: "
        	+Long.toString(SystemProperties.getLong("nw.cache.trimamount", CACHE_TRIM_AMOUNT_DEF)));
        Log.d("speculative", "CONNECTION_COUNT: "
        	+Integer.toString(SystemProperties.getInt("http.threads", 4)));
    }
    // runnable for loading resource metadata
    public static class LoadSaveResourceMetadataRunnable implements Runnable {
    	// true: load
    	// false: save
    	private boolean mIsLoad;
    	public LoadSaveResourceMetadataRunnable(boolean flag) {
    		mIsLoad = flag;
    	}
    	public synchronized void run() {
    		// load resource metadata from file; get called when enabling speculative loading
    		if (mIsLoad) {
    			Log.d("speculative", "load resource metadata ===================================================");
				FileInputStream fis = null;
				ObjectInputStream in = null;
				try {
					fis = new FileInputStream(mResourceMetadataFileName);
					in = new ObjectInputStream(fis);
					mResourceMetadata = (ArrayList<ResourceGraph>)in.readObject();
					// set to clean
					mDirtyResourceMetadata = false;
					Log.d("speculative", "CacheManager loadResourceMetadata: load from resource metadata file sucessfully");
				}
				catch (FileNotFoundException ex) {
					Log.d("speculative", "CacheManager loadResourceMetadata: no saved resource metadata file");
					mResourceMetadata = new ArrayList<ResourceGraph>();
				}
				catch (IOException ex) {
					Log.d("speculative", "CacheManager loadResourceMetadata: IOException!!!! "+ex.getMessage());
					ex.printStackTrace();
				}
				catch(ClassNotFoundException ex) {
					Log.d("speculative", "CacheManager loadResourceMetadata: ClassNotFoundException!!!! "+ex.getMessage());
					ex.printStackTrace();
				}
				Log.d("speculative", "resource metadata loaded ===================================================");
				dumpResourceMetadata();
				// enable speculation here, resource metadata is ready now
				mSpeculativeEnabled = true;
				Log.d("speculative", "\n*********************************************\n"
									+"*********** speculation enabled *************\n"
									+"*********************************************\n\n");
			}
    		// save resource metadata to file
			else {
				Log.d("speculative", "save resource metadata ===================================================");
				if (mResourceMetadata != null) {
					FileOutputStream fos = null;
					ObjectOutputStream out = null;
					try {
						fos = new FileOutputStream(mResourceMetadataFileName);
						out = new ObjectOutputStream(fos);
						out.writeObject(mResourceMetadata);
						out.close();
						// set to clean
						mDirtyResourceMetadata = false;
						Log.d("speculative", "CacheManager loadResourceMetadata: save to resource metadata file sucessfully");
					}
					catch (IOException ex) {
						Log.d("speculative", "CacheManager saveResourceMetadata: IOException!!!! "+ex.getMessage());
						ex.printStackTrace();
					}
				}
				else {
					Log.d("speculative", "CacheManager saveResourceMetadata: mResourceMetadata == null - Don't save!");
				}
				Log.d("speculative", "resource metadata saved ===================================================");
				dumpResourceMetadata();
			}
    	}
    }
    // called by onResume() in BrowserActivity.java
    /**
     * @hide
     */
    public static void loadResourceMetadata() {
    	Thread loadThread = new Thread(null, new LoadSaveResourceMetadataRunnable(true), "loadThread", 256*1024);
    	loadThread.start();
    }
    // called by onPause() in BrowserActivity.java
    /**
     * @hide
     */
    public static void saveResourceMetadata() {
    	Thread saveThread = new Thread(null, new LoadSaveResourceMetadataRunnable(false), "saveThread", 256*1024);
    	saveThread.start();
    }
    
    private static synchronized ResourceGraph getResourceGraph(String site) {
    	ResourceGraph resource_graph = null;
    	for (int i=0; i<mResourceMetadata.size(); i++) {
    		resource_graph = mResourceMetadata.get(i);
    		if (site.equals(resource_graph.getSite())) {
    			return resource_graph;
    		}
    	}
    	return null;
    }
    
    /**
     * @hide
     */
    public static synchronized void updateResourceMetadata() {
    	if (mResourceMetadata == null) {
    		Log.d("speculative", "CacheManager updateResourceMetadata: mResourceMetadata is NOT ready !!!!");
    		return;
    	}
    	if (mCurrentWebpage == null) {
    		Log.d("speculative", "CacheManager updateResourceMetadata: mCurrentWebpage == null !!!!");
    		return;
    	}
    	// get the subresource array
    	String[] subresources = new String[mCurrentSubresources.size()];
    	for (int i=0; i<mCurrentSubresources.size(); i++) {
    		subresources[i] = mCurrentSubresources.get(i);
    	}
    	// update
    	long currentTime = System.currentTimeMillis();
    	String site = URLUtil.url2site(mCurrentWebpage);
    	ResourceGraph resource_graph = getResourceGraph(site);
    	if (resource_graph != null) {
    		resource_graph.update(mCurrentWebpage, subresources, currentTime);
    	}
    	else {
    		resource_graph = new ResourceGraph(site, currentTime);
    		resource_graph.update(mCurrentWebpage, subresources, currentTime);
    		mResourceMetadata.add(resource_graph);
    	}
    	Log.d("speculative", "CacheManager updateResourceMetadata for "+mCurrentWebpage);
    	for (int i=0; i<subresources.length; i++) {
    		Log.d("speculative", " - "+subresources[i]);
    	}
    	// set resource metadata to dirty
    	mDirtyResourceMetadata = true;
    	// clear current webpage and subresources before next visit to capture
    	// subres requested by AJAX
    }
    
    /**
     * @hide
     */
    public static synchronized String[] predictFromResourceMetadata(String webpage) {
    	if (mResourceMetadata == null) {
    		Log.d("speculative", "CacheManager predictFromResourceMetadata: mResourceMetadata is NOT ready !!!!");
    		return null;
    	}
    	long currentTime = System.currentTimeMillis();
    	String site = URLUtil.url2site(webpage);
    	ResourceGraph resource_graph = getResourceGraph(site);
    	Log.d("speculative", "CacheManager predictFromResourceMetadata for ("+site+") "+webpage);
    	if (resource_graph != null) {
    		return resource_graph.predict(webpage);
    	}
    	else {
    		return null;
    	}
    }
    
    /**
     * @hide
     */
    public static synchronized void trimResourceMetadata() {
    	if (mResourceMetadata == null) {
    		Log.d("speculative", "CacheManager trimResourceMetadata: mResourceMetadata is NOT ready !!!!");
    		return;
    	}
    	long currentTime = System.currentTimeMillis();
    	ResourceGraph resource_graph = null;
    	for (int i=0; i<mResourceMetadata.size(); i++) {
    		resource_graph = mResourceMetadata.get(i);
    		resource_graph.trim(currentTime);
    	}
    	Log.d("speculative", "CacheManager trimResourceMetadata");
    	// set resource metadata to dirty
    	mDirtyResourceMetadata = true;
    }
    
    private static String mSpeculativeLog = null;
    private static String mSpeculativeLogFileName = "/sdcard/speculative/log";
    private static boolean mIsFirstSubresRequested = true;
    public static boolean isFirstSubresRequested() {
    	return mIsFirstSubresRequested;
    }
    public static void setNotFirstSubresRequested() {
    	mIsFirstSubresRequested = false;
    }
    public static void setFirstSubresRequested() {
    	mIsFirstSubresRequested = true;
    }
    public static synchronized void appendSpeculativeLog(String line) {
    	if (mSpeculativeLog==null) {
    		mSpeculativeLog = "\n\n"+line+'\n';
    	}
    	else {
    		mSpeculativeLog += line+'\n';
    	}
    }
    // only called when saving the record, already synchronized
    private static void recordSubresToSpeculativeLog() {
    	// record speculative links
   		mSpeculativeLog += "numPredictedTotal "+Integer.toString(mPredictedQueue.size())+'\n';
   		mSpeculativeLog += "numPredictedLoad "+Integer.toString(mSpeculativeFinishedPool.size()
   																+mSpeculativeErrorPool.size())+'\n';
   		mSpeculativeLog += "numActualLoad "+Integer.toString(mCurrentSubresources.size())+'\n';
   		for (int i=0; i<mPredictedQueue.size(); i++) {
   			mSpeculativeLog += "predictedSubres "+mPredictedQueue.get(i)+'\n';
   		}
   		for (int i=0; i<mCurrentSubresources.size(); i++) {
   			mSpeculativeLog += "actualSubres "+mCurrentSubresources.get(i)+'\n';
   		}        
    }
    public static synchronized void saveSpeculativeLogOnPageFinished(String url) {
    	try {
    		if (mSpeculativeLog != null) {
    			// first time to save the log, also record subres
    			mSpeculativeLog += "onPageFinished "+Long.toString(System.currentTimeMillis())+" "+url+'\n';
    			if (speculativeEnabled()) {
    				recordSubresToSpeculativeLog();
    			}
    		}
    		else {
    			mSpeculativeLog = "onPageFinished "+Long.toString(System.currentTimeMillis())+" "+url+'\n';
    		}
    		BufferedWriter out = new BufferedWriter(new FileWriter(mSpeculativeLogFileName, true));
    		out.write(mSpeculativeLog);
    		out.close();
    		//reset
    		mSpeculativeLog = null;
    		mIsFirstSubresRequested = false;
    	} catch (IOException ex) {
    		Log.d("speculative", "CacheManager saveSpeculativeLog IOException: "+ex.getMessage());
    		ex.printStackTrace();
    	}
    }
    // this can be called after onPageFinished
    public static synchronized void saveSpeculativeLogOnProgress100() {
    	try {
    		if (mSpeculativeLog != null) {
    			// first time to save the log, also record subres
    			mSpeculativeLog += "SetProgress100 "+Long.toString(System.currentTimeMillis())+'\n';
    			if (speculativeEnabled()) {
    				recordSubresToSpeculativeLog();
    			}
    		}
    		else {
    			mSpeculativeLog = "SetProgress100 "+Long.toString(System.currentTimeMillis())+'\n';
    		}
    		BufferedWriter out = new BufferedWriter(new FileWriter(mSpeculativeLogFileName, true));
    		out.write(mSpeculativeLog);
    		out.close();
    		//reset
    		mSpeculativeLog = null;
    		mIsFirstSubresRequested = false;
    	} catch (IOException ex) {
    		Log.d("speculative", "CacheManager setProgress100SpeculativeLog IOException: "+ex.getMessage());
    		ex.printStackTrace();
    	}
    }
    
    private static boolean mPageStarted = false;
    private static boolean mPageFinished = true;
    public static void onPageStarted(String url) {
    	// there can be multiple onPageStarted if having redictions
    	// only act at the 1st one
    	if (mPageStarted==false) {
			// update previous resource metadata and clear previous record
			updateResourceMetadata();
			clearCurrentWebpageSubresources();
			// set the flags
			mPageStarted = true;
			mPageFinished = false;
			// record the new webpage if not google redirect url
			if (!URLUtil.isGoogleUrlRedirect(url)) {
				setCurrentWebpage(url);
			}
    	}
    	// if on 1st page started, the url is google redirect url, webpage will
    	// not be set. So set it now
    	else if (mCurrentWebpage == null) {
			setCurrentWebpage(url);
    	}
    }
    public static void onPageFinished() {
    	// update the resource metadata
		updateResourceMetadata();
		// set the flags
    	mPageStarted = false;
    	mPageFinished = true;
    	// don't clear the record yet; so we can capture subframe and AJAX requests
    	// resource metadata will be updated again before next onPageStarted
    }
}
