/*
 * Copyright 2007, The Android Open Source Project
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
#include "config.h"
#include "InspectorControllerAndroid.h"
#include "InspectorClient.h"

#include "Frame.h"
#include "Node.h"
#if USE(JSC)
#include "Profile.h"
#endif

#include "Event.h"
#include "IntRect.h"
#include "ResourceRequest.h"
#include "ResourceResponse.h"

#include <wtf/CurrentTime.h>

#define LOG_TAG "TimeLine"
#include <utils/Log.h>

namespace WebCore {
	
InspectorController::InspectorController(Page*, InspectorClient* client)
	: m_client(client)
    , m_page(0)
{
	LOGD("%s", "InspectorController constructor =============================");
    m_client = client;
    startTimelineProfiler();
}

InspectorController::~InspectorController() { 
	LOGD("%s", "InspectorController destructor =============================");
    stopTimelineProfiler();
    m_client->inspectorDestroyed(); 
}

void InspectorController::inspectedPageDestroyed() 
{
    m_client->inspectorDestroyed();
    m_client = 0;
}

bool InspectorController::enabled() const { return true; }

// willSendRequest
void InspectorController::willSendRequest(unsigned long identifier, ResourceRequest& request, ResourceResponse const& redirectResponse) 
{
    if (m_timelineAgent)
        m_timelineAgent->willSendResourceRequest(identifier, request);
}
// didFinishLoading
void InspectorController::didFinishLoading(unsigned long identifier) 
{
    if (m_timelineAgent)
        m_timelineAgent->didFinishLoadingResource(identifier, false);
}
// didFailLoading
void InspectorController::didFailLoading(unsigned long identifier, const ResourceError& )
{
    if (m_timelineAgent)
        m_timelineAgent->didFinishLoadingResource(identifier, true);
}

// Timeline Controller: for Backend
void InspectorController::startTimelineProfiler() // backend
{
    if(!enabled())
        return;
    LOGD("%s", "startTimelineProfiler =======================================");
    if (m_timelineAgent)
        return;
    m_timelineAgent = new InspectorTimelineAgent();
}

void InspectorController::stopTimelineProfiler() // backend
{
    if(!enabled())
        return;
    if (!m_timelineAgent)
        return;
    m_timelineAgent.clear();
}

/*
bool InspectorController::windowVisible() { return false; }
void InspectorController::windowScriptObjectAvailable() {}
void InspectorController::didCommitLoad(DocumentLoader*) {}
void InspectorController::identifierForInitialRequest(unsigned long, DocumentLoader*, ResourceRequest const&) {}
void InspectorController::didReceiveResponse(DocumentLoader*, unsigned long, ResourceResponse const&) {}
void InspectorController::didReceiveContentLength(DocumentLoader*, unsigned long, int) {}
void InspectorController::didLoadResourceFromMemoryCache(DocumentLoader*, const CachedResource*) {}
void InspectorController::frameDetachedFromParent(Frame*) {}
void InspectorController::addMessageToConsole(WebCore::MessageSource, WebCore::MessageType, WebCore::MessageLevel, WebCore::String const&, unsigned int, WebCore::String const&) {}
void InspectorController::addMessageToConsole(WebCore::MessageSource, WebCore::MessageType, WebCore::MessageLevel, ScriptCallStack*) {}
void InspectorController::inspect(Node*) {}
void InspectorController::resourceRetrievedByXMLHttpRequest(unsigned long identifier, const ScriptString& sourceString) {}
void InspectorController::scriptImported(unsigned long identifier, const String& sourceString) {}
void InspectorController::inspectedWindowScriptObjectCleared(Frame* frame) {}
void InspectorController::startGroup(MessageSource source, ScriptCallStack* callFrame) {}
void InspectorController::endGroup(MessageSource source, unsigned lineNumber, const String& sourceURL) {}
void InspectorController::startTiming(const String& title) {}
bool InspectorController::stopTiming(const String& title, double& elapsed) { return false; }
void InspectorController::count(const String& title, unsigned lineNumber, const String& sourceID) {}
void InspectorController::mouseDidMoveOverElement(const HitTestResult&, unsigned modifierFlags) {}
void InspectorController::handleMousePressOnNode(Node*) {}
*/


// Timeline Agent =======================================================================================
int InspectorTimelineAgent::RecordLevel = 0;

InspectorTimelineAgent::InspectorTimelineAgent()
{
    startTime_global = 0;
    timelineOverhead = 0;
}

InspectorTimelineAgent::~InspectorTimelineAgent()
{
}

void InspectorTimelineAgent::willCallFunction(const String& scriptName, int scriptLine)
{
    pushCurrentRecord(TimelineRecordFactory::createFunctionCallData(scriptName, scriptLine), FunctionCallTimelineRecordType);
}

void InspectorTimelineAgent::didCallFunction()
{
    didCompleteCurrentRecord(FunctionCallTimelineRecordType);
}

void InspectorTimelineAgent::willDispatchEvent(const Event& event)
{
    pushCurrentRecord(TimelineRecordFactory::createEventDispatchData(event),
        EventDispatchTimelineRecordType);
}

void InspectorTimelineAgent::didDispatchEvent()
{
    didCompleteCurrentRecord(EventDispatchTimelineRecordType);
}

void InspectorTimelineAgent::willLayout()
{
    pushCurrentRecord(newTimelineObject(), LayoutTimelineRecordType);
}

void InspectorTimelineAgent::didLayout()
{
    didCompleteCurrentRecord(LayoutTimelineRecordType);
}

void InspectorTimelineAgent::willRecalculateStyle()
{
    pushCurrentRecord(newTimelineObject(), RecalculateStylesTimelineRecordType);
}

void InspectorTimelineAgent::didRecalculateStyle()
{
    didCompleteCurrentRecord(RecalculateStylesTimelineRecordType);
}

void InspectorTimelineAgent::willPaint(const IntRect& rect)
{
    pushCurrentRecord(TimelineRecordFactory::createPaintData(rect), PaintTimelineRecordType);
}

void InspectorTimelineAgent::didPaint()
{
    didCompleteCurrentRecord(PaintTimelineRecordType);
}

void InspectorTimelineAgent::willRender()
{
    pushCurrentRecord(newTimelineObject(), RenderTimelineRecordType);
}

void InspectorTimelineAgent::didRender()
{
    didCompleteCurrentRecord(RenderTimelineRecordType);
}

void InspectorTimelineAgent::willWriteHTML(unsigned int length, unsigned int startLine, unsigned long identifier)
{
    pushCurrentRecord(TimelineRecordFactory::createParseHTMLData(length, startLine, identifier), ParseHTMLTimelineRecordType);
}

void InspectorTimelineAgent::didWriteHTML(unsigned int endLine)
{
    if (!m_recordStack.isEmpty()) {
        TimelineRecordEntry record = m_recordStack.last();
        record.data.endLine = endLine;
        didCompleteCurrentRecord(ParseHTMLTimelineRecordType);
    }
}

void InspectorTimelineAgent::didInstallTimer(int timerId, int timeout, bool singleShot)
{
    TimelineObject data = TimelineRecordFactory::createTimerInstallData(timerId, timeout, singleShot);
    data.startTime = WTF::currentTime() * 1000.0;
    TimelineRecordEntry record = TimelineRecordEntry(data, newTimelineArray(), TimerInstallTimelineRecordType);
    addRecordToTimeline(record, TimerInstallTimelineRecordType);
}

void InspectorTimelineAgent::didRemoveTimer(int timerId)
{
    TimelineObject data = TimelineRecordFactory::createGenericTimerData(timerId);
    data.startTime = WTF::currentTime() * 1000.0;
    TimelineRecordEntry record = TimelineRecordEntry(data, newTimelineArray(), TimerRemoveTimelineRecordType);
    addRecordToTimeline(record, TimerRemoveTimelineRecordType);
}

void InspectorTimelineAgent::willFireTimer(int timerId)
{
    pushCurrentRecord(TimelineRecordFactory::createGenericTimerData(timerId), TimerFireTimelineRecordType); 
}

void InspectorTimelineAgent::didFireTimer()
{
    didCompleteCurrentRecord(TimerFireTimelineRecordType);
}

void InspectorTimelineAgent::willChangeXHRReadyState(const String& url, int readyState)
{
    pushCurrentRecord(TimelineRecordFactory::createXHRReadyStateChangeData(url, readyState), XHRReadyStateChangeRecordType);
}

void InspectorTimelineAgent::didChangeXHRReadyState()
{
    didCompleteCurrentRecord(XHRReadyStateChangeRecordType);
}

void InspectorTimelineAgent::willLoadXHR(const String& url) 
{
    pushCurrentRecord(TimelineRecordFactory::createXHRLoadData(url), XHRLoadRecordType);
}

void InspectorTimelineAgent::didLoadXHR()
{
    didCompleteCurrentRecord(XHRLoadRecordType);
}

void InspectorTimelineAgent::willEvaluateScript(const String& url, int lineNumber)
{
    pushCurrentRecord(TimelineRecordFactory::createEvaluateScriptData(url, lineNumber), EvaluateScriptTimelineRecordType);
}
    
void InspectorTimelineAgent::didEvaluateScript()
{
    didCompleteCurrentRecord(EvaluateScriptTimelineRecordType);
}

void InspectorTimelineAgent::didScheduleResourceRequest(const String& url)
{
    TimelineObject data = TimelineRecordFactory::createScheduleResourceRequestData(url);
    data.startTime = WTF::currentTime() * 1000.0;
    TimelineRecordEntry record = TimelineRecordEntry(data, newTimelineArray(), ScheduleResourceRequestTimelineRecordType);
    addRecordToTimeline(record, ScheduleResourceRequestTimelineRecordType);
}

void InspectorTimelineAgent::willSendResourceRequest(unsigned long identifier,
    const ResourceRequest& request)
{
    TimelineObject data = TimelineRecordFactory::createResourceSendRequestData(identifier, request);
    data.startTime = WTF::currentTime() * 1000.0;
    TimelineRecordEntry record = TimelineRecordEntry(data, newTimelineArray(), ResourceSendRequestTimelineRecordType);
    saveRecord(record);
}

void InspectorTimelineAgent::willReceiveResourceData(unsigned long identifier)
{
    pushCurrentRecord(TimelineRecordFactory::createReceiveResourceData(identifier), ReceiveResourceDataTimelineRecordType);
}

void InspectorTimelineAgent::didReceiveResourceData()
{
    didCompleteCurrentRecord(ReceiveResourceDataTimelineRecordType);
}
    
void InspectorTimelineAgent::willReceiveResourceResponse(unsigned long identifier, const ResourceResponse& response)
{
    pushCurrentRecord(TimelineRecordFactory::createResourceReceiveResponseData(identifier, response), ResourceReceiveResponseTimelineRecordType);
}

void InspectorTimelineAgent::didReceiveResourceResponse()
{
    didCompleteCurrentRecord(ResourceReceiveResponseTimelineRecordType);
}

void InspectorTimelineAgent::didFinishLoadingResource(unsigned long identifier, bool didFail)
{
    TimelineObject data = TimelineRecordFactory::createResourceFinishData(identifier, didFail);
    data.startTime = WTF::currentTime() * 1000.0;
    TimelineRecordEntry record = TimelineRecordEntry(data, newTimelineArray(), ResourceFinishTimelineRecordType);
    saveRecord(record);
}

void InspectorTimelineAgent::didMarkTimeline(const String& message)
{
    TimelineObject data = TimelineRecordFactory::createMarkTimelineData(message);
    data.startTime = WTF::currentTime() * 1000.0;
    TimelineRecordEntry record = TimelineRecordEntry(data, newTimelineArray(), MarkTimelineRecordType);
    addRecordToTimeline(record, MarkTimelineRecordType);
}

void InspectorTimelineAgent::didMarkDOMContentEvent()
{
    TimelineObject data = TimelineRecordFactory::createGenericRecord((WTF::currentTime()*1000.0));
    TimelineRecordEntry record = TimelineRecordEntry(data, newTimelineArray(), MarkDOMContentEventType);
    addRecordToTimeline(record, MarkDOMContentEventType);
}

void InspectorTimelineAgent::didMarkLoadEvent()
{  
    TimelineObject data = TimelineRecordFactory::createGenericRecord((WTF::currentTime()*1000.0));
    TimelineRecordEntry record = TimelineRecordEntry(data, newTimelineArray(), MarkLoadEventType);
    addRecordToTimeline(record, MarkLoadEventType);
}

void InspectorTimelineAgent::reset()
{
    m_recordStack.clear();
    m_timelineRecords.clear();
    RecordLevel = 0;
    startTime_global = WTF::currentTime()*1000.0;
    timelineOverhead = 0;
}

void InspectorTimelineAgent::addRecordToTimeline(TimelineRecordEntry record, TimelineRecordType type)
{
    record.type = type;
    if (m_recordStack.isEmpty()) {
        saveRecord(record);
    }
    else {
        TimelineRecordEntry parent = m_recordStack.last();
        m_recordStack.removeLast();
        parent.children.append(record);
        m_recordStack.append(parent);
    }
}

void InspectorTimelineAgent::didCompleteCurrentRecord(TimelineRecordType type)
{
    // An empty stack could merely mean that the timeline agent was turned on in the middle of
    // an event.  Don't treat as an error.
    if (!m_recordStack.isEmpty()) {
        TimelineRecordEntry record = m_recordStack.last();
        m_recordStack.removeLast();
        ASSERT(record.type == type);
        record.data.endTime = WTF::currentTime() * 1000.0;
        addRecordToTimeline(record, type);
    }
}

void InspectorTimelineAgent::pushCurrentRecord(TimelineObject data, TimelineRecordType type)
{
    data.startTime = WTF::currentTime() * 1000.0;
    m_recordStack.append(TimelineRecordEntry(data, newTimelineArray(), type));
}

static const char* TimelineTypeNames[] = {
	"EventDispatchTimelineRecordType",
    "Layout",
    "RecalculateStyles",
    "Paint",
    "ParseHTML",
	"TimerInstallTimelineRecordType",
	"TimerRemoveTimelineRecordType",
	"TimerFireTimelineRecordType",
	"XHRReadyStateChangeRecordType",
	"XHRLoadRecordType",
    "EvaluateScript",
	"MarkTimelineRecordType",
    "ResourceSendRequest",
    "ResourceReceiveResponse",
    "ResourceFinish",
	"FunctionCallTimelineRecordType",
    "ReceiveResourceData",
	"GCEventTimelineRecordType",
	"MarkDOMContentEventType",
	"MarkLoadEventType",
    "ScheduleResourceRequest",
    "Render",
};

void InspectorTimelineAgent::saveRecord(TimelineRecordEntry record)
{
    m_timelineRecords.append(record);
}

void InspectorTimelineAgent::sortTimelineRecord(Vector<TimelineRecordEntry>& records)
{
	TimelineRecordEntry tmp = TimelineRecordEntry(TimelineObject::createNew(), newTimelineArray(), TempType);
    for (size_t j = 0; j < (records.size()-1); ++j) {
        for (size_t i = 0; i < (records.size()-j-1); ++i){
            if (records[i].data.startTime > records[i+1].data.startTime){
                tmp = records[i];
                records[i] = records[i+1];
                records[i+1] = tmp;
            }
        }
    }
	// sort children
	for (size_t k = 0; k < records.size(); ++k) {
		if (!records[k].children.isEmpty()) 
			sortTimelineRecord(records[k].children);
	}
}

void InspectorTimelineAgent::writeRecordToSDcard()
{
    // sort m_timelineRecords by start time; bubble sort, may be slow
	sortTimelineRecord(m_timelineRecords);
    
    FILE* TimelineFile = fopen("/sdcard/timeline_log.txt", "a");
    // write the time reference:
    fprintf(TimelineFile, "%s%f\n", "##START_TIME: ", startTime_global);
    //fprintf(TimelineFile, "%s\n", "##***********************************************************************");
    // write to file
    for (Vector<TimelineRecordEntry>::iterator it = m_timelineRecords.begin(); it != m_timelineRecords.end(); ++it) {
        writeRecordToFile(TimelineFile, *it);
    }
    fclose(TimelineFile);
    
    // clear the record
    m_recordStack.clear();
    m_timelineRecords.clear();
    RecordLevel = 0;
}

void InspectorTimelineAgent::writeRecordToFile(FILE* TimelineFile, TimelineRecordEntry record)
{
    double start = record.data.startTime - startTime_global;
    double end = record.data.endTime - startTime_global;
    
    switch (record.type) {
        case ScheduleResourceRequestTimelineRecordType:
            fprintf(TimelineFile, "%s\t%f\t%s\n", TimelineTypeNames[record.type], 
                start, record.data.url.utf8().data()); 
            break;
        case ResourceSendRequestTimelineRecordType:
            fprintf(TimelineFile, "%s\t(%lu)\t%f\t%s\n", TimelineTypeNames[record.type], 
                record.data.identifier, start, record.data.url.utf8().data()); 
            break;
        case ResourceFinishTimelineRecordType:
            fprintf(TimelineFile, "%s\t(%lu)\t%f\n", TimelineTypeNames[record.type], 
                record.data.identifier, start); 
            break;
        case ResourceReceiveResponseTimelineRecordType:
        case ReceiveResourceDataTimelineRecordType:
        case ParseHTMLTimelineRecordType:
            fprintf(TimelineFile, "%s\t(%lu)\t%f\t%f\t%f\n", TimelineTypeNames[record.type], 
                record.data.identifier, start, end, end - start); 
            break;
        case EvaluateScriptTimelineRecordType:
            fprintf(TimelineFile, "%s\t%f\t%f\t%f\t%s\n", TimelineTypeNames[record.type], 
                start, end, end - start, record.data.url.utf8().data());
            break;
        default: // Layout, RecalculateStyles, Paint, Render
            fprintf(TimelineFile, "%s\t%f\t%f\t%f\n", TimelineTypeNames[record.type], 
                start, end, end - start); 
    }
        
    // print children
    for (Vector<TimelineRecordEntry>::iterator it = record.children.begin(); it != record.children.end(); ++it) {
        ++RecordLevel;
        for(int i = 0; i < RecordLevel; ++i)
            fprintf(TimelineFile, "%s", "---> ");
        writeRecordToFile(TimelineFile, *it);
        --RecordLevel;
    }
}

TimelineObject InspectorTimelineAgent::newTimelineObject()
{
    return TimelineObject::createNew();
}

Vector<TimelineRecordEntry> InspectorTimelineAgent::newTimelineArray()
{
    Vector<TimelineRecordEntry> TimelineArray;
    return TimelineArray;
}

// Timeline Record Factory ====================================================================================
TimelineObject TimelineRecordFactory::createGenericRecord(double startTime)
{
    TimelineObject record = InspectorTimelineAgent::newTimelineObject();
    record.startTime = startTime;
    return record;
}

TimelineObject TimelineRecordFactory::createFunctionCallData(const String& scriptName, int scriptLine)
{
    TimelineObject data = InspectorTimelineAgent::newTimelineObject();
    data.scriptName = scriptName;
    data.scriptLine = scriptLine;
    return data;
}

TimelineObject TimelineRecordFactory::createEventDispatchData(const Event& event)
{
    TimelineObject data = InspectorTimelineAgent::newTimelineObject();
    data.type = event.type().string();
    return data;
}

TimelineObject TimelineRecordFactory::createGenericTimerData(int timerId)
{
    TimelineObject data = InspectorTimelineAgent::newTimelineObject();
    data.timerId = timerId;
    return data;
}

TimelineObject TimelineRecordFactory::createTimerInstallData(int timerId, int timeout, bool singleShot)
{
    TimelineObject data = InspectorTimelineAgent::newTimelineObject();
    data.timerId = timerId;
    data.timeout = timeout;
    data.singleShot = singleShot;
    return data;
}

TimelineObject TimelineRecordFactory::createXHRReadyStateChangeData(const String& url, int readyState)
{
    TimelineObject data = InspectorTimelineAgent::newTimelineObject();
    data.url = url;
    data.readyState = readyState;
    return data;
}

TimelineObject TimelineRecordFactory::createXHRLoadData(const String& url)
{
    TimelineObject data = InspectorTimelineAgent::newTimelineObject();
    data.url = url;
    return data;
}

TimelineObject TimelineRecordFactory::createEvaluateScriptData(const String& url, double lineNumber) 
{
    TimelineObject data = InspectorTimelineAgent::newTimelineObject();
    data.url = url;
    data.lineNumber = lineNumber;
    return data;
}

TimelineObject TimelineRecordFactory::createMarkTimelineData(const String& message) 
{
    TimelineObject data = InspectorTimelineAgent::newTimelineObject();
    data.message = message;
    return data;
}

TimelineObject TimelineRecordFactory::createScheduleResourceRequestData(const String& url)
{
    TimelineObject data = InspectorTimelineAgent::newTimelineObject();
    data.url = url;
    return data;
}

TimelineObject TimelineRecordFactory::createResourceSendRequestData(unsigned long identifier, const ResourceRequest& request)
{
    TimelineObject data = InspectorTimelineAgent::newTimelineObject();
    data.identifier = identifier;
    data.url = request.url().string();
    data.requestMethod = request.httpMethod();
    return data;
}

TimelineObject TimelineRecordFactory::createResourceReceiveResponseData(unsigned long identifier, const ResourceResponse& response)
{
    TimelineObject data = InspectorTimelineAgent::newTimelineObject();
    data.identifier = identifier;
    data.statusCode = response.httpStatusCode();
    data.mimeType = response.mimeType();
    data.expectedContentLength = response.expectedContentLength();
    return data;
}

TimelineObject TimelineRecordFactory::createResourceFinishData(unsigned long identifier, bool didFail)
{
    TimelineObject data = InspectorTimelineAgent::newTimelineObject();
    data.identifier = identifier;
    data.didFail = didFail;
    return data;
}

TimelineObject TimelineRecordFactory::createReceiveResourceData(unsigned long identifier)
{
    TimelineObject data = InspectorTimelineAgent::newTimelineObject();
    data.identifier = identifier;
    return data;
}
    
TimelineObject TimelineRecordFactory::createPaintData(const IntRect& rect)
{
    TimelineObject data = InspectorTimelineAgent::newTimelineObject();
    data.x = rect.x();
    data.y = rect.y();
    data.width = rect.width();
    data.height = rect.height();
    return data;
}

TimelineObject TimelineRecordFactory::createParseHTMLData(unsigned int length, unsigned int startLine, unsigned long identifier)
{
    TimelineObject data = InspectorTimelineAgent::newTimelineObject();
    data.length = length;
    data.startLine = startLine;
    data.identifier = identifier;
    return data;
}

}  // namespace WebCore
