#ifndef InspectorControllerAndroid_h
#define InspectorControllerAndroid_h

#include "PlatformString.h"
#include "ScriptState.h"
#include "StringHash.h"
#include "Timer.h"

#include <wtf/HashMap.h>
#include <wtf/HashSet.h>
#include <wtf/RefCounted.h>
#include <wtf/Vector.h>

#include "Document.h"
#include "ScriptObject.h"
#include "ScriptArray.h"
#include <wtf/Vector.h>

#include <stdio.h>
#include "CString.h"

namespace WebCore {

class CachedResource;
class Document;
class DocumentLoader;
class HitTestResult;
class InspectorClient;
class KURL;
class Node;
class Page;
struct ResourceRequest;
class ResourceResponse;
class ResourceError;
class ScriptCallStack;
class ScriptObject;
class ScriptString;

class Event;
class IntRect;

class InspectorTimelineAgent;

class InspectorController {
public:
    OwnPtr<InspectorTimelineAgent> m_timelineAgent;
    InspectorTimelineAgent* timelineAgent() { return m_timelineAgent.get(); }
    void startTimelineProfiler();
    void stopTimelineProfiler();

    InspectorController(Page*, InspectorClient*);
    ~InspectorController();

    void pageDestroyed() { m_page = 0; }
    void inspectedPageDestroyed();

    bool enabled() const;

    void willSendRequest(unsigned long identifier, ResourceRequest& request, ResourceResponse const& redirectResponse);
    void didFinishLoading(unsigned long identifier);
    void didFailLoading(unsigned long identifier, const ResourceError& error);

    // not actually defined
/*    bool windowVisible();
    void windowScriptObjectAvailable();
    void didCommitLoad(DocumentLoader*);
    void identifierForInitialRequest(unsigned long identifier, DocumentLoader*, const ResourceRequest&);
    void didReceiveResponse(DocumentLoader*, unsigned long identifier, const ResourceResponse&);
    void didReceiveContentLength(DocumentLoader*, unsigned long identifier, int lengthReceived);
    void didLoadResourceFromMemoryCache(DocumentLoader*, const CachedResource*);
    void frameDetachedFromParent(Frame*);
    void addMessageToConsole(MessageSource, MessageType, MessageLevel, ScriptCallStack*);
    void addMessageToConsole(MessageSource, MessageType, MessageLevel, const String& message, unsigned lineNumber, const String& sourceID);
    void inspect(Node*);
    void resourceRetrievedByXMLHttpRequest(unsigned long identifier, const ScriptString& sourceString);
    void scriptImported(unsigned long identifier, const String& sourceString);
    void inspectedWindowScriptObjectCleared(Frame*);
    void startGroup(MessageSource source, ScriptCallStack* callFrame);
    void endGroup(MessageSource source, unsigned lineNumber, const String& sourceURL);
    void startTiming(const String& title);
    bool stopTiming(const String& title, double& elapsed);
    void count(const String& title, unsigned lineNumber, const String& sourceID);
    void mouseDidMoveOverElement(const HitTestResult&, unsigned modifierFlags);
    void handleMousePressOnNode(Node*);*/


private:
    InspectorClient* m_client;
    Page* m_page;
};


enum TimelineRecordType {
    EventDispatchTimelineRecordType = 0,
    LayoutTimelineRecordType = 1,
    RecalculateStylesTimelineRecordType = 2,
    PaintTimelineRecordType = 3,
    ParseHTMLTimelineRecordType = 4,
    TimerInstallTimelineRecordType = 5,
    TimerRemoveTimelineRecordType = 6,
    TimerFireTimelineRecordType = 7,
    XHRReadyStateChangeRecordType = 8,
    XHRLoadRecordType = 9,
    EvaluateScriptTimelineRecordType = 10,
    MarkTimelineRecordType = 11,
    ResourceSendRequestTimelineRecordType = 12,
    ResourceReceiveResponseTimelineRecordType = 13,
    ResourceFinishTimelineRecordType = 14,
    FunctionCallTimelineRecordType = 15,
    ReceiveResourceDataTimelineRecordType = 16,
    GCEventTimelineRecordType = 17,
    MarkDOMContentEventType = 18,
    MarkLoadEventType = 19,
    ScheduleResourceRequestTimelineRecordType = 20,
    RenderTimelineRecordType = 21,
    TempType = 22
};

// Timeline Object ------------------------------------------------------
class TimelineObject {
    public:
        TimelineObject() {url="Null";}
        static TimelineObject createNew() {return TimelineObject();};
        
        // generic
        double startTime;
        double endTime;
        String url;
        int readState;
        double lineNumber;
        
        // function call
        String scriptName;
        int scriptLine;
        
        // event dispatch 
        String type;
        
        // timer
        int timerId;
        int timeout;
        bool singleShot;
        
        // XHR Ready State Change
        int readyState;
        
        // mark timeline
        String message;
        
        // resource request, response & finish
        unsigned long identifier;
        String requestMethod;
        bool isMainResource;
        int statusCode;
        String mimeType;
        long long expectedContentLength;
        bool didFail;
        
        //paint
        int x;
        int y;
        int width;
        int height;
        
        // parse html
        unsigned int length;
        unsigned int startLine;
        unsigned int endLine;
};

// Timeline Record Entry ----------------------------------------
struct TimelineRecordEntry {
        TimelineRecordEntry(TimelineObject data, Vector<TimelineRecordEntry> children, TimelineRecordType type)
            : data(data), children(children), type(type)
        {
        }
        TimelineObject data;
        Vector<TimelineRecordEntry> children;
        TimelineRecordType type;
    };
    
// Timeline Agent ---------------------------------------------------
class InspectorTimelineAgent : public Noncopyable {
public:
    InspectorTimelineAgent();
    ~InspectorTimelineAgent();

    void reset();

    // Methods called from WebCore.
    void willCallFunction(const String& scriptName, int scriptLine);
    void didCallFunction();

    void willDispatchEvent(const Event&);
    void didDispatchEvent();

    void willLayout();
    void didLayout();

    void willRecalculateStyle();
    void didRecalculateStyle();

    void willPaint(const IntRect&);
    void didPaint();
    
    void willRender();
    void didRender();

    void willWriteHTML(unsigned int length, unsigned int startLine, unsigned long identifier);
    void didWriteHTML(unsigned int endLine);

    void didInstallTimer(int timerId, int timeout, bool singleShot);
    void didRemoveTimer(int timerId);
    void willFireTimer(int timerId);
    void didFireTimer();

    void willChangeXHRReadyState(const String&, int);
    void didChangeXHRReadyState();
    void willLoadXHR(const String&);
    void didLoadXHR();

    void willEvaluateScript(const String&, int);
    void didEvaluateScript();

    void didMarkTimeline(const String&);
    void didMarkDOMContentEvent();
    void didMarkLoadEvent();

    void didScheduleResourceRequest(const String& url);
    void willSendResourceRequest(unsigned long, const ResourceRequest&);

    void willReceiveResourceResponse(unsigned long, const ResourceResponse&);
    void didReceiveResourceResponse();

    void willReceiveResourceData(unsigned long identifier);
    void didReceiveResourceData();

    void didFinishLoadingResource(unsigned long, bool didFail);

    static InspectorTimelineAgent* retrieve(ScriptExecutionContext*);

    void saveRecord(TimelineRecordEntry record); //m_frontend->addRecordToTimeline(record);
    void sortTimelineRecord(Vector<TimelineRecordEntry>& records);
    void writeRecordToSDcard(); 
    void writeRecordToFile(FILE* TimelineFile, TimelineRecordEntry record);
    static int RecordLevel;
    static TimelineObject newTimelineObject(); //m_frontend->newScriptObject();
    static Vector<TimelineRecordEntry> newTimelineArray(); //m_frontend->newScriptArray();
    
private:
    void pushCurrentRecord(TimelineObject, TimelineRecordType);
        
    void didCompleteCurrentRecord(TimelineRecordType);

    void addRecordToTimeline(TimelineRecordEntry, TimelineRecordType);

    Vector<TimelineRecordEntry> m_timelineRecords;
    Vector<TimelineRecordEntry> m_recordStack;
    
public: 
    double startTime_global;
    double timelineOverhead;
};

/* XXX This is needed in eclair to get timeline agent instance, may have new ways in gingerbread
inline InspectorTimelineAgent* InspectorTimelineAgent::retrieve(ScriptExecutionContext* context)
{
    if (context && context->isDocument())
        return static_cast<Document*>(context)->inspectorTimelineAgent();
    return 0;
}
*/

// Timeline Record Factory -----------------------------------------------
class TimelineRecordFactory {
    public:
        static TimelineObject createGenericRecord(double startTime);

        static TimelineObject createFunctionCallData(const String& scriptName, int scriptLine);

        static TimelineObject createEventDispatchData(const Event&);

        static TimelineObject createGenericTimerData(int timerId);

        static TimelineObject createTimerInstallData(int timerId, int timeout, bool singleShot);

        static TimelineObject createXHRReadyStateChangeData(const String& url, int readyState);

        static TimelineObject createXHRLoadData(const String& url);

        static TimelineObject createEvaluateScriptData(const String& url, double lineNumber);

        static TimelineObject createMarkTimelineData(const String&);

        static TimelineObject createResourceSendRequestData(unsigned long identifier, const ResourceRequest&);

        static TimelineObject createScheduleResourceRequestData(const String&);

        static TimelineObject createResourceReceiveResponseData(unsigned long identifier, const ResourceResponse&);

        static TimelineObject createReceiveResourceData(unsigned long identifier);

        static TimelineObject createResourceFinishData(unsigned long identifier, bool didFail);

        static TimelineObject createPaintData(const IntRect&);

        static TimelineObject createParseHTMLData(unsigned int length, unsigned int startLine, unsigned long identifier);

    private:
        TimelineRecordFactory() { }
};

} // namespace WebCore

#endif // !defined(InspectorControllerAndroid_h)
